<?php

namespace Nest\Cli;

use Nest\Application;
use Nest\Cli\Commands\CommandContext;
use Nest\Cli\Commands\CommandDefinition;
use Nest\Cli\Commands\FlagDefinition;
use Nest\Cli\Exceptions\Command\CommandException;
use Nest\Cli\Exceptions\Command\InvalidCommandHandlerException;
use Nest\Cli\Exceptions\IncompatibleCliHandlerSubcommands;
use Throwable;

/**
 * CLI manager and engine.
 */
class Cli
{
	/**
	 * Associative array of CLI commands.
	 * @var array<string, CommandDefinition>
	 */
	protected array $commands;

	/**
	 * The CLI helper.
	 * @var CliHelper
	 */
	protected CliHelper $helper;

	/**
	 * @param Application $application The application.
	 * @param CliConfiguration $configuration CLI configuration.
	 */
	public function __construct(protected Application $application, protected CliConfiguration $configuration)
	{
	}

	/**
	 * Create a new command context for given arguments.
	 * @param string[]|null $argv Command arguments, NULL to get the currently passed command arguments.
	 * @return CommandContext The command context.
	 * @throws CommandException
	 */
	public function newCommandContext(?array $argv = null): CommandContext
	{
		// Initialize command context with arguments.
		return new CommandContext($this, $this->commands, $argv);
	}

	/**
	 * Exception handler for the CLI.
	 * @param Throwable $exception Exception to handle.
	 * @return void
	 */
	protected function exceptionHandler(Throwable $exception): void
	{
		Out::error("Unhandled exception: {$exception->getMessage()}");
		Out::error("in file ".Out::ITALIC_ATTRIBUTE."{$exception->getFile()}".Out::RESET_ATTRIBUTES.Color::Red->value." on line {$exception->getLine()}.");
		echo Color::Red->value."{$exception->getTraceAsString()}".Color::Default->value."\n";
	}

	/**
	 * Main engine function.
	 * @return void
	 * @throws InvalidCommandHandlerException
	 * @throws IncompatibleCliHandlerSubcommands
	 */
	public function __invoke(): void
	{
		// Add exception handler.
		$this->application->exceptionHandler()->add(fn (Throwable $exception) => $this->exceptionHandler($exception));

		// Define configuration of this CLI.
		$this->configuration->define($this);

		// Get CLI helper.
		$helper = $this->getHelper();

		try
		{
			// Create and fill command context, then check its validity.
			$commandContext = $this->newCommandContext();
			$commandContext->checkValidity();

			if ($commandContext->hasFlag(FlagDefinition::HELP_FLAG))
			{ // If auto help flag has been provided, showing help no matter the rest.
				$helper->help($this, $commandContext);
			}
			else
			{ // Execute the command handler with all parameters.
				$commandContext->getCommandHandler()(...$commandContext->getParameters());
			}
		}
		catch (CommandException $commandException)
		{ // If any command parse error happen, show it, then show help for the given context.
			$helper->help($this, $commandException->commandContext, $commandException);
		}
	}

	/**
	 * Set CLI helper.
	 * The role of CLI helper is to provide help when needed (either asked by user or after invalid arguments).
	 * @param CliHelper $helper New CLI helper.
	 * @return void
	 */
	public function setHelper(CliHelper $helper): void
	{
		$this->helper = $helper;
	}

	/**
	 * Get CLI helper.
	 * If none is set, return a default one.
	 * @return CliHelper CLI helper instance.
	 */
	public function getHelper(): CliHelper
	{
		// Return defined helper, if there is one.
		if (!empty($this->helper)) return $this->helper;

		// Return a new default helper.
		return new DefaultHelper();
	}

	/**
	 * Get command definition of the given command.
	 * Initialize a new command definition if it didn't exist.
	 * @param string $command A command
	 * @return CommandDefinition
	 */
	public function command(string $command): CommandDefinition
	{
		if (empty($this->commands[$command]))
			// Initialize a new command definition because it doesn't exist.
			$this->commands[$command] = new CommandDefinition();
		// Return command definition of given command.
		return $this->commands[$command];
	}

	/**
	 * Get commands definitions for the CLI.
	 * @return array<string, CommandDefinition> Commands definitions.
	 */
	public function getCommands(): array
	{
		return $this->commands;
	}
}
