<?php

namespace Nest\Cli\Commands;

use Nest\Cli\Cli;
use Nest\Cli\Exceptions\Command\CommandException;
use Nest\Cli\Exceptions\Command\IncompleteCommandException;
use Nest\Cli\Exceptions\Command\InvalidCommandException;
use Nest\Cli\Exceptions\Command\NotEnoughParametersException;
use Nest\Cli\Exceptions\Command\MissingCommandException;
use Nest\Cli\Exceptions\Command\MissingRequiredFlagException;
use Nest\Cli\Exceptions\Command\UndefinedFlagException;
use Nest\Cli\Exceptions\Command\UndefinedShortFlagException;
use Nest\Cli\Exceptions\Command\UndefinedSubcommandsException;
use Nest\Cli\Exceptions\FlagNotFoundException;

/**
 * Class of a command context.
 */
class CommandContext
{
	/**
	 * Command executable.
	 * @var string
	 */
	protected string $executable;

	/**
	 * Commands path (with all subcommands).
	 * @var string[]
	 */
	protected array $commandsPath;

	/**
	 * Associated command definition.
	 * @var CommandDefinition
	 */
	protected CommandDefinition $commandDefinition;

	/**
	 * Command parameters.
	 * @var string[]
	 */
	protected array $parameters = [];

	/**
	 * Flags with their parameters (if there are some).
	 * @var array<string, array>
	 */
	protected array $flags;

	/**
	 * Command handler callable, determined from command definition.
	 * @var callable(mixed...): ?int
	 */
	protected $handler;

	/**
	 * Create a new command context and fill it using passed command arguments.
	 * @param Cli $cli The CLI.
	 * @param array<string, CommandDefinition> $commands Commands definitions.
	 * @param array|null $argv Command arguments (with executable at first position), NULL to get the currently passed command arguments.
	 * @throws CommandException
	 */
	public function __construct(protected Cli $cli, protected array $commands, ?array $argv = null)
	{
		// Parse command.
		$this->parseCommand($argv);
	}

	/**
	 * Parse passed command arguments.
	 * @param array|null $args Command arguments, NULL to get the currently passed command arguments.
	 * @return void
	 * @throws CommandException
	 */
	protected function parseCommand(?array $args = null): void
	{
		// Get command arguments.
		global $argv;

		// If args are not defined, get it from argv.
		if (is_null($args)) $args = $argv;

		// Save used executable.
		$this->executable = $args[0];

		if (empty($args[1]))
			// No command.
			throw new MissingCommandException($this);

		// Parse commands path from first argument.
		$this->commandsPath = array_map(fn (string $command) => trim($command), explode(":", $args[1]));

		// Find associated command definition from commands path.
		$this->findCommandDefinition();

		if (!empty($args[2]))
			// Find command flags and parameters, if there are some.
			$this->findCommandFlagsAndParams(array_slice($args, 2));
	}

	/**
	 * Find associated command definition from commands path.
	 * @param int $currentIndex Current index in commands path.
	 * @param CommandDefinition|null $currentCommandDefinition Current command definition, for recursive calls.
	 * @return void
	 * @throws CommandException
	 */
	protected function findCommandDefinition(int $currentIndex = 0, ?CommandDefinition $currentCommandDefinition = null): void
	{
		if (empty($this->commandsPath[$currentIndex]))
		{ // We reached the end of the commands path, set the current command definition as the current one.
			$this->commandDefinition = $currentCommandDefinition;

			if ($this->commandDefinition->hasHandler())
			{ // Check that a command handler is defined for the current command.
				// Get command handler.
				$commandHandler = $this->commandDefinition->getHandler();
				// Instantiate handler class to call the object.
				$this->handler = new $commandHandler($this->cli, $this);
			}
			else
				// If there is no handler for the current command definition, the given command is incomplete.
				throw new IncompleteCommandException($this);
		}
		else
		{ // Get the next command definition, if there is one.

			// Get current command part.
			$commandPart = $this->commandsPath[$currentIndex];

			if (empty($currentCommandDefinition))
			{ // Try to read root command.
				if (empty($this->commands[$commandPart]))
					// Invalid command.
					throw new InvalidCommandException($commandPart, $this);

				// Recursive call to find the right command definition.
				$this->findCommandDefinition($currentIndex + 1, $this->commands[$commandPart]);
			}
			else
			{ // Try to read a subcommand.
				if ($currentCommandDefinition->hasSubcommands())
				{ // Try to get subcommands.
					$subcommands = $currentCommandDefinition->getSubcommands();

					if (is_array($subcommands))
					{ // If it's an array, try to get the right command definition.
						if (empty($subcommands[$commandPart]))
							// Invalid command.
							throw new InvalidCommandException($commandPart, $this);

						// Recursive call to find the right command definition.
						$this->findCommandDefinition($currentIndex + 1, $subcommands[$commandPart]);
					}
					else
					{ // A subcommands handler is defined, we just need to call it with the right subcommand.
						$subcommandsHandler = $currentCommandDefinition->getSubcommands();

						// Instantiate handler class to call the object.
						$callable = new $subcommandsHandler($this->cli, $this);

						// Get remaining subcommands.
						$subcommandsPath = array_slice($this->commandsPath, $currentIndex);

						// Set current command definition as definitive command definition.
						$this->commandDefinition = $currentCommandDefinition;
						// Create a new handler which will pass subcommands path as first parameter of the callable.
						$this->handler = function (mixed ...$parameters) use ($callable, $subcommandsPath) {
							return $callable($subcommandsPath, ...$parameters);
						};
					}
				}
				else
					// No subcommands for the current command, throwing an exception.
					throw new UndefinedSubcommandsException($commandPart, $this);
			}
		}
	}

	/**
	 * Find command flags and parameters.
	 * @param string[] $args Raw arguments array.
	 * @return void
	 * @throws UndefinedFlagException
	 */
	protected function findCommandFlagsAndParams(array $args): void
	{
		// Initialize current flag (full) name.
		$currentFlagName = null;

		foreach ($args as $arg)
		{ // Reading each argument one by one, to determine if it's a flag, a flag parameter or a positional parameter.

			if (str_starts_with($arg, "--"))
			{ // Current argument is a flag.
				// Get current flag name.
				$currentFlagName = substr($arg, 2);

				if (empty($this->commandDefinition->getFlagDefinition($currentFlagName)) && $currentFlagName != FlagDefinition::HELP_FLAG)
					// If the flag is undefined, throw an exception. Explicitly allow auto help flag.
					throw new UndefinedFlagException($currentFlagName, $this);

				// Initialize flag in the context.
				$this->flags[$currentFlagName] = [];
			}
			elseif (str_starts_with($arg, "-"))
			{ // Current argument is a (short) flag.
				// Get current flag name from short flag name.
				$currentFlagName = $this->commandDefinition->getShortFlagName($shortFlagName = substr($arg, 1));

				if (empty($currentFlagName))
					// If the short flag is undefined, throw an exception.
					throw new UndefinedShortFlagException($shortFlagName, $this);

				// Initialize flag in the context.
				$this->flags[$currentFlagName] = [];
			}
			else
			{ // Current argument is a parameter.
				if (!empty($currentFlagName) && count($this->flags[$currentFlagName]) < $this->commandDefinition->getFlagDefinition($currentFlagName)->getParametersCount())
				{ // If there is a current flag and we did not read all its parameters, considering this parameter is a flag parameter.
					$this->flags[$currentFlagName][] = $arg;
				}
				else
					// No current flag (no current flag or no remaining parameters for it), the parameter is a command parameter.
					$this->parameters[] = $arg;
			}
		}
	}

	/**
	 * Check command validity: the current command context should fulfill requirements of the command definition.
	 * Throw a CommandException if anything is wrong.
	 * @return void
	 * @throws CommandException
	 */
	public function checkValidity(): void
	{
		// Check there are enough parameters to match the required count.
		if (($providedCount = !empty($this->parameters) ? count($this->parameters) : 0) < ($expectedCount = $this->commandDefinition->getRequiredParametersCount()))
			throw new NotEnoughParametersException($providedCount, $expectedCount, $this);

		// Check that all the required flags are set.
		foreach ($this->commandDefinition->getRequiredFlags() as $requiredFlag)
		{ // Check each required flag existence.
			if (!$this->hasFlag($requiredFlag))
				// Required flag is not defined, throw an exception.
				throw new MissingRequiredFlagException($requiredFlag, $this);
		}
	}

	/**
	 * Get used executable.
	 * @return string Command executable.
	 */
	public function getExecutable(): string
	{
		return $this->executable;
	}

	/**
	 * Get full command (with all its subcommands).
	 * @return string Full command.
	 */
	public function getFullCommand(): string
	{
		return implode(":", $this->commandsPath);
	}

	/**
	 * Get commands path (all subcommands).
	 * @return string[]
	 */
	public function getCommandsPath(): array
	{
		return $this->commandsPath;
	}

	/**
	 * Get associated command definition, or NULL if it couldn't be found.
	 * @return CommandDefinition|null
	 */
	public function getCommandDefinition(): ?CommandDefinition
	{
		return $this->commandDefinition ?? null;
	}

	/**
	 * Get command handler callable.
	 * @return callable(mixed...): ?int
	 */
	public function getCommandHandler(): callable
	{
		return $this->handler;
	}

	/**
	 * Get command positional parameters.
	 * @return array
	 */
	public function getParameters(): array
	{
		return $this->parameters;
	}

	/**
	 * Determine if a flag has been provided in the command.
	 * @param string $flag Flag (full) name.
	 * @return bool True if the flag have been provided, false otherwise.
	 */
	public function hasFlag(string $flag): bool
	{
		return isset($this->flags[$flag]);
	}

	/**
	 * Get flag parameters of a flag, if they were provided.
	 * @param string $flag Flag (full) name.
	 * @return array Flag parameters.
	 * @throws FlagNotFoundException
	 */
	public function getFlagParameters(string $flag): array
	{
		if (!$this->hasFlag($flag))
			// Flag not provided, throw an exception.
			throw new FlagNotFoundException($flag);

		return $this->flags[$flag]; // Return flag parameters.
	}
}
