<?php

namespace Nest\Cli\Commands;

use Nest\Cli\Exceptions\Command\InvalidCommandHandlerException;
use Nest\Cli\Exceptions\IncompatibleCliHandlerSubcommands;

/**
 * A command definition object.
 */
class CommandDefinition
{
	/**
	 * The command description.
	 * @var string|null
	 */
	protected ?string $description = null;

	/**
	 * Parameters array for the current command.
	 * @var array<string, ParameterDefinition>
	 */
	protected array $parameters = [];

	/**
	 * Flags array for the current command.
	 * @var array<string, FlagDefinition>
	 */
	protected array $flags = [];

	/**
	 * Short flags matching array.
	 * Associate every short flag with its matching full flag.
	 * @var array<string, string>
	 */
	protected array $shortFlags = [];

	/**
	 * Command handler.
	 * @var string
	 */
	protected string $handler;

	/**
	 * Subcommands handler or subcommands definitions.
	 * @var string|array<string, CommandDefinition>
	 */
	protected string|array $subcommands;

	/**
	 * Set the description of the command.
	 * @param string $description The description of the command.
	 * @return $this
	 */
	public function description(string $description): static
	{
		// Set description.
		$this->description = $description;
		return $this;
	}

	/**
	 * Define a parameter for the command.
	 * @param string $name Parameter name.
	 * @param (callable(ParameterDefinition): ParameterDefinition)|null $parameterDefinition Parameter definition function.
	 * @return $this
	 */
	public function parameter(string $name, ?callable $parameterDefinition = null): static
	{
		// Add the given parameter with its definition.
		$this->parameters[$name] = new ParameterDefinition();

		if (!empty($parameterDefinition))
			// A parameter definition function is defined, calling it.
			$this->parameters[$name] = $parameterDefinition($this->parameters[$name]);

		return $this;
	}

	/**
	 * Define a flag for the command.
	 * @param string $flag Flag name.
	 * @param (callable(FlagDefinition): FlagDefinition)|null $flagDefinition Flag definition function.
	 * @return $this
	 */
	public function flag(string $flag, ?callable $flagDefinition = null): static
	{
		// Add the given flag with its definition.
		$this->flags[$flag] = new FlagDefinition();

		if (!empty($flagDefinition))
			// A flag definition function is defined, calling it.
			$this->flags[$flag] = $flagDefinition($this->flags[$flag]);

		if (!empty($short = $this->flags[$flag]->getShort()))
			// If a short flag is defined, add it to short flag matching array.
			$this->shortFlags[$short] = $flag;

		return $this;
	}

	/**
	 * Define an handler for the command.
	 * Cannot be used when subcommands are defined.
	 * @param string $handler Command handler class.
	 * @return $this
	 * @throws IncompatibleCliHandlerSubcommands
	 * @throws InvalidCommandHandlerException
	 */
	public function handler(string $handler): static
	{
		if (!empty($this->subcommands))
			// Subcommands are already defined, throwing an exception.
			throw new IncompatibleCliHandlerSubcommands();

		// Check that the command handler has the right type.
		if (!is_a($handler, CommandHandler::class, true))
			throw new InvalidCommandHandlerException($handler);

		// Set the current command handler.
		$this->handler = $handler;

		return $this;
	}

	/**
	 * Define subcommands for the command.
	 * Cannot be used when handler is defined.
	 * @param string|array<string, callable(CommandDefinition): CommandDefinition> $subcommands Subcommands handler or subcommands definitions array.
	 * @return $this
	 * @throws IncompatibleCliHandlerSubcommands
	 * @throws InvalidCommandHandlerException
	 */
	public function subcommands(string|array $subcommands): static
	{
		if (!empty($this->handler))
			// A handler is already defined, throwing an exception.
			throw new IncompatibleCliHandlerSubcommands();

		if (is_string($subcommands))
		{ // If subcommands is a handler class, check that the subcommands handler has the right type.
			if (!is_a($subcommands, CommandHandler::class, true))
				throw new InvalidCommandHandlerException($subcommands);
			// Set subcommands handler.
			$this->subcommands = $subcommands;
		}
		else
		{ // Set subcommands definition array.
			foreach ($subcommands as $subcommand => $definitionCallable)
			{ // For each subcommand definition, call its definition function.
				$this->subcommands[$subcommand] = $definitionCallable(new CommandDefinition());
			}
		}

		return $this;
	}

	/**
	 * Return true if a command handler is defined for this command.
	 * @return bool
	 */
	public function hasHandler(): bool
	{
		return !empty($this->handler);
	}

	/**
	 * Get defined command handler class.
	 * @return string Command handler class.
	 */
	public function getHandler(): string
	{
		return $this->handler;
	}

	/**
	 * Return true if subcommands are defined for this command.
	 * @return bool
	 */
	public function hasSubcommands(): bool
	{
		return !empty($this->subcommands);
	}

	/**
	 * Get subcommands definition.
	 * @return string|array<string, callable(CommandDefinition): CommandDefinition> Subcommands handler or subcommands definitions.
	 */
	public function getSubcommands(): string|array
	{
		return $this->subcommands;
	}

	/**
	 * Get required parameters count.
	 * @return int Required parameters count.
	 */
	public function getRequiredParametersCount(): int
	{
		// Check every parameter to count the required ones.
		return array_reduce(
			$this->parameters,
			// Add 1 everytime we see a required parameter.
			fn (int $total, ParameterDefinition $parameter) => $total + ($parameter->isRequired() ? 1 : 0),
			0,
		);
	}

	/**
	 * Get command description.
	 * @return string
	 */
	public function getDescription(): string
	{
		return $this->description ?? "";
	}

	/**
	 * Get defined parameters for the command.
	 * @return array<string, ParameterDefinition> Defined parameters.
	 */
	public function getParameters(): array
	{
		return $this->parameters;
	}

	/**
	 * Get short flag full name.
	 * @param string $shortFlag Short flag name.
	 * @return string|null Flag full name. NULL if undefined.
	 */
	public function getShortFlagName(string $shortFlag): ?string
	{
		return $this->shortFlags[$shortFlag] ?? null;
	}

	/**
	 * Get flag definition for the given flag (full) name.
	 * @param string $flag Flag (full) name.
	 * @return FlagDefinition|null Flag definition. NULL if undefined.
	 */
	public function getFlagDefinition(string $flag): ?FlagDefinition
	{
		return $this->flags[$flag] ?? null;
	}

	/**
	 * Get required flags list.
	 * @return string[] Required flags names.
	 */
	public function getRequiredFlags(): array
	{
		// Initialize required flags list.
		$requiredFlags = [];

		foreach ($this->flags as $flag => $definition)
		{ // For each flag, if it is required, add it in the list.
			if ($definition->isRequired())
				// The current flag is required, adding it to the list.
				$requiredFlags[] = $flag;
		}

		return $requiredFlags;
	}

	/**
	 * Get defined flags for the command.
	 * @return array<string, FlagDefinition> Defined flags.
	 */
	public function getFlags(): array
	{
		return $this->flags;
	}
}
