<?php

namespace Nest\Cli;

use Nest\Cli\Commands\CommandContext;
use Nest\Cli\Commands\CommandDefinition;
use Throwable;

/**
 * Default CLI helper.
 */
class DefaultHelper implements CliHelper
{
	public const string REQUIRED_FIELD_PREFIX = "{";
	public const string REQUIRED_FIELD_SUFFIX = "}";

	public const string OPTIONAL_FIELD_PREFIX = "[";
	public const string OPTIONAL_FIELD_SUFFIX = "]";

	/**
	 * The command context.
	 * @var CommandContext
	 */
	private CommandContext $command;

	/**
	 * @inheritDoc
	 */
	#[\Override] public function help(Cli $cli, CommandContext $command, ?Throwable $exception = null): void
	{
		// Save command context in class property.
		$this->command = $command;

		if (!empty($exception))
			// If an exception has been thrown, show its message.
			Out::error($exception->getMessage());

		echo "\n";

		if (!empty($commandDefinition = $command->getCommandDefinition()))
		{ // A command definition have been provided. Showing help for this command especially.
			$this->showHelp($command->getFullCommand(), $commandDefinition, true);
		}
		else
		{
			echo Out::BOLD_ATTRIBUTE."Available commands: ".Out::RESET_ATTRIBUTES."\n\n";
			foreach ($cli->getCommands() as $command => $commandDefinition)
			{ // Showing help for each command.
				// Dimmed list tick.
				echo Out::DIM_ATTRIBUTE;
				echo " -  ";
				echo Out::RESET_ATTRIBUTES;
				// Command help.
				$this->showHelp($command, $commandDefinition);
			}
		}
	}

	/**
	 * Show help for a specific command definition.
	 * @param string $command Current command to show.
	 * @param CommandDefinition $commandDefinition Full command definition.
	 * @param bool $details If true, help will show detailed information about parameters and flags.
	 * @return void
	 */
	private function showHelp(string $command, CommandDefinition $commandDefinition, bool $details = false): void
	{
		if ($details)
		{ // Details requested.
			if ($commandDefinition->hasSubcommands())
			{ // If there are subcommands, showing all subcommands.
				$subcommands = $commandDefinition->getSubcommands();
				if (is_array($subcommands))
				{ // Showing subcommands array details.
					echo Out::BOLD_ATTRIBUTE." Subcommands: ".Out::RESET_ATTRIBUTES."\n\n";

					foreach ($subcommands as $subcommand => $subcommandDefinition)
					{
						// Dimmed list tick.
						echo Out::DIM_ATTRIBUTE;
						echo " -  ";
						echo Out::RESET_ATTRIBUTES;
						// Show subcommand help.
						$this->showHelp("$command:$subcommand", $subcommandDefinition);
					}
				}
				else
				{ // Showing generic subcommands manual.
					echo "{$this->command->getExecutable()} $command";
					echo ":{}";
					echo "\n\n";
					echo Out::DIM_ATTRIBUTE." Customized subcommands are not documented.".Out::RESET_ATTRIBUTES;
				}
			}
			elseif ($commandDefinition->hasHandler())
			{ // If there is an handler, showing all parameters with details.
				echo "{$this->command->getExecutable()} $command";

				$this->printParameters($commandDefinition);
				$this->printFlags($commandDefinition);

				echo "\n\n";

				if (!empty($commandDefinition->getParameters()))
				{ // Show full parameters details.
					echo Out::BOLD_ATTRIBUTE." Parameters: ".Out::RESET_ATTRIBUTES."\n\n";
					foreach ($commandDefinition->getParameters() as $parameter => $parameterDefinition)
					{ // Show details of each parameter.

						// Dimmed list tick.
						echo Out::DIM_ATTRIBUTE;
						if (!$parameterDefinition->isRequired())
							echo "  *  (optional) ";
						else
							echo "  -  ";
						echo Out::RESET_ATTRIBUTES;

						echo "$parameter: ".Out::ITALIC_ATTRIBUTE."{$parameterDefinition->getDescription()}".Out::RESET_ATTRIBUTES;
						echo "\n";
					}

					// Parameters end.
					echo "\n";
				}
				else
					echo Out::DIM_ATTRIBUTE." No parameters.\n".Out::RESET_ATTRIBUTES;

				if (!empty($commandDefinition->getFlags()))
				{ // Show full flags details.
					// Show flags in yellow.
					echo Color::LightYellow->value;

					// Show full flags details.
					echo Out::BOLD_ATTRIBUTE." Flags: ".Out::RESET_ATTRIBUTES.Color::LightYellow->value."\n";

					foreach ($commandDefinition->getFlags() as $flag => $flagDefinition)
					{ // Show details of each parameter.

						echo "\n";

						$flagColor = $flagDefinition->isRequired() ? Color::LightRed->value : Color::LightYellow->value;

						// Dimmed list tick.
						echo $flagColor.Out::DIM_ATTRIBUTE;
						echo "  -  ";
						if ($flagDefinition->isRequired()) echo "(required) ";
						echo Out::RESET_ATTRIBUTES.$flagColor;

						echo "--$flag";
						if (!empty($short = $flagDefinition->getShort()))
						{ // If a short equivalent is defined, showing it.
							echo " (alt.: -$short)";
						}
						echo ": ".Out::ITALIC_ATTRIBUTE."{$flagDefinition->getDescription()}".Out::RESET_ATTRIBUTES.Color::LightYellow->value;
					}

					// Flags end.
					echo Color::Default->value;
				}
				else
					echo Out::DIM_ATTRIBUTE." No flags.\n".Out::RESET_ATTRIBUTES;
			}
			else
			{ // Something is wrong with the definition.
				echo "\n";
				Out::error("Invalid command definition.");
			}
		}
		else
		{ // Details not requested, simply show parameters and flags.
			echo "{$this->command->getExecutable()} $command";

			$this->printParameters($commandDefinition);
			$this->printFlags($commandDefinition);

			if (!empty($description = $commandDefinition->getDescription()))
			{
				echo "\n";
				echo Out::ITALIC_ATTRIBUTE."    $description".Out::RESET_ATTRIBUTES;
			}
		}

		echo "\n\n";
	}

	/**
	 * Print parameters of a command.
	 * @param CommandDefinition $commandDefinition Command definition.
	 * @return void
	 */
	private function printParameters(CommandDefinition $commandDefinition): void
	{
		foreach ($commandDefinition->getParameters() as $parameter => $parameterDefinition)
		{ // Show every parameter according to its definition.
			echo " ";
			echo $parameterDefinition->isRequired() ? static::REQUIRED_FIELD_PREFIX : static::OPTIONAL_FIELD_PREFIX;
			echo $parameter;
			echo $parameterDefinition->isRequired() ? static::REQUIRED_FIELD_SUFFIX : static::OPTIONAL_FIELD_SUFFIX;
		}
	}

	/**
	 * Print flags of a command.
	 * @param CommandDefinition $commandDefinition Command definition.
	 * @return void
	 */
	private function printFlags(CommandDefinition $commandDefinition): void
	{
		foreach ($commandDefinition->getFlags() as $flag => $flagDefinition)
		{ // Show every flag according to its definition.
			echo " ";
			echo $flagDefinition->isRequired() ? static::REQUIRED_FIELD_PREFIX : static::OPTIONAL_FIELD_PREFIX;
			echo "--$flag";
			echo $flagDefinition->isRequired() ? static::REQUIRED_FIELD_SUFFIX : static::OPTIONAL_FIELD_SUFFIX;
		}
	}
}
